{ Main view, where most of the application logic takes place.

  Feel free to use this code as a starting point for your own projects.
  This template code is in public domain, unlike most other CGE code which
  is covered by BSD or LGPL (see https://castle-engine.io/license). }
unit GameViewMain;

interface

uses Classes,
  CastleVectors, CastleComponentSerialize, CastleScene, CastleTransform,
  CastleUIControls, CastleControls, CastleKeysMouse, CastleImages, CastleTerrain;

type
  { Main view, where most of the application logic takes place. }
  TViewMain = class(TCastleView)
  published
    { Components designed using CGE editor.
      These fields will be automatically initialized at Start. }
    LabelFps: TCastleLabel;
    TreeImage: TCastleImageTransform;
    Scene1,Scene2,Scene3,Scene4,Scene5,
    Scene6,Scene7,Scene8,Scene9,Scene10: TCastleScene;
    Plane1: TCastlePlane;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Start; override;
    procedure Update(const SecondsPassed: Single; var HandleInput: Boolean); override;
    function Press(const Event: TInputPressRelease): Boolean; override;
  private
    type
      TDiggiTestCase = (diggiTestImages, diggiTestPlanes, diggiTestScenes, diggiTestClones);
  private
    procedure AddImages(Original: TCastleImageTransform; distanceMultiplier, ANumber: Integer; blend: Boolean);
    procedure AddPlanes(Original: TCastlePlane; distanceMultiplier, ANumber: Integer; blend: Boolean);
    procedure AddScenes(Original: TCastleScene; distanceMultiplier, ANumber: Integer; blend: Boolean);
    procedure AddClones(Original: TCastleScene; distanceMultiplier, ANumber: Integer; blend: Boolean);
    procedure Visibility_Images(AVisible: Boolean; blend: Boolean);
    procedure Visibility_Planes(AVisible: Boolean; blend: Boolean);
    procedure Visibility_Scenes(AVisible: Boolean; blend: Boolean);
    procedure TestImages(distanceMultiplier, ANumber: Integer; blend: Boolean);
    procedure TestPlanes(distanceMultiplier, ANumber: Integer; blend: Boolean);
    procedure TestScenes(distanceMultiplier, ANumber: Integer; blend: Boolean);
    procedure TestClones(distanceMultiplier, ANumber: Integer; blend: Boolean);
  end;

var
  ViewMain: TViewMain;

implementation

uses SysUtils;

{ TViewMain ----------------------------------------------------------------- }

procedure TViewMain.AddClones(Original: TCastleScene; distanceMultiplier, ANumber: Integer; blend: Boolean);
var
  Clone: TCastleTransformReference;
  i: Integer;
begin
  for i := 1 to ANumber do
  begin
    Clone := TCastleTransformReference.Create(Original.Owner);
    Clone.Reference := Original;
    Clone.Direction := Original.Direction;
    Clone.Translation := Vector3(distanceMultiplier*Random(5000), 0, -distanceMultiplier*Random(5000));
    Clone.Scale := Original.Scale;

    Original.Parent.Add(Clone);
  end;
end;

procedure TViewMain.AddImages(Original: TCastleImageTransform; distanceMultiplier, ANumber: Integer; blend: Boolean);
var
  Img: TCastleImageTransform;
  i: Integer;
begin
  for i := 1 to ANumber do
  begin
    Img := TCastleImageTransform.Create(Original.Owner);
    Img.Direction := Original.Direction;
    Img.Translation := Vector3(distanceMultiplier*Random(5000), 492, -distanceMultiplier*Random(5000));
    Img.Scale := Original.Scale;
    if blend then
      Img.AlphaChannel := acBlending
    else
      Img.AlphaChannel := acNone; //acTest;
    Img.RenderOptions.Blending := blend;

    Img.Url := Original.Url;
    Original.Parent.Add(Img);
  end;
end;

procedure TViewMain.AddPlanes(Original: TCastlePlane; distanceMultiplier, ANumber: Integer; blend: Boolean);
var
  Plane: TCastlePlane;
  i: Integer;
begin
  for i := 1 to ANumber do
  begin
    Plane := TCastlePlane.Create(Original.Owner);
    Plane.Color := Original.Color;
    Plane.Direction := Original.Direction;
    Plane.Translation := Vector3(distanceMultiplier*Random(5000), 0, -distanceMultiplier*Random(5000));
    Plane.Size := Original.Size;
    Plane.Scale := Original.Scale;
    Plane.RenderOptions.Blending := blend;

    Plane.Texture := Original.Texture;
    Plane.TextureNormalMap := Original.TextureNormalMap;

    Original.Parent.Add(Plane);
  end;
end;

procedure TViewMain.AddScenes(Original: TCastleScene; distanceMultiplier, ANumber: Integer; blend: Boolean);
var
  Model: TCastleScene;
  i: Integer;
begin
  for i := 1 to ANumber do
  begin
    Model := TCastleScene.Create(Original.Owner);
    Model.Cache := true;
    Model.Direction := Original.Direction;
    Model.Translation := Vector3(distanceMultiplier*Random(5000), 0, -distanceMultiplier*Random(5000));
    Model.Scale := Original.Scale; //Vector3(1, 1, 1);
    Model.RenderOptions.Blending := blend;

    Model.Url := Original.Url;
    Original.Parent.Add(Model);
  end;
end;

constructor TViewMain.Create(AOwner: TComponent);
begin
  inherited;
  DesignUrl := 'castle-data:/gameviewmain.castle-user-interface';
end;

procedure TViewMain.Visibility_Images(AVisible, blend: Boolean);
begin
  TreeImage.Exists := AVisible;
  TreeImage.RenderOptions.Blending := blend;
  if blend then
    TreeImage.AlphaChannel := acBlending
  else
    TreeImage.AlphaChannel := acNone; //acTest;
end;

procedure TViewMain.Visibility_Planes(AVisible, blend: Boolean);
begin
  Plane1.Exists := AVisible;
  Plane1.RenderOptions.Blending := blend;
end;

procedure TViewMain.Visibility_Scenes(AVisible, blend: Boolean);
begin
  Scene1.Exists := AVisible;
  Scene2.Exists := AVisible;
  Scene3.Exists := AVisible;
  Scene4.Exists := AVisible;
  Scene5.Exists := AVisible;
  Scene6.Exists := AVisible;
  Scene7.Exists := AVisible;
  Scene8.Exists := AVisible;
  Scene9.Exists := AVisible;
  Scene10.Exists := AVisible;

  Scene1.RenderOptions.Blending := blend;
  Scene2.RenderOptions.Blending := blend;
  Scene3.RenderOptions.Blending := blend;
  Scene4.RenderOptions.Blending := blend;
  Scene5.RenderOptions.Blending := blend;
  Scene6.RenderOptions.Blending := blend;
  Scene7.RenderOptions.Blending := blend;
  Scene8.RenderOptions.Blending := blend;
  Scene9.RenderOptions.Blending := blend;
  Scene10.RenderOptions.Blending := blend;
end;

procedure TViewMain.Start;
begin
  inherited;
end;

procedure TViewMain.TestClones(distanceMultiplier, ANumber: Integer; blend: Boolean);
var
  n: Integer;
begin
  Visibility_Images(false, blend);
  Visibility_Planes(false, blend);
  Visibility_Scenes(true, blend);

  n := (ANumber div 10) - 1;
  AddClones(Scene1, distanceMultiplier,n, blend);
  AddClones(Scene2, distanceMultiplier,n, blend);
  AddClones(Scene3, distanceMultiplier,n, blend);
  AddClones(Scene4, distanceMultiplier,n, blend);
  AddClones(Scene5, distanceMultiplier,n, blend);
  AddClones(Scene6, distanceMultiplier,n, blend);
  AddClones(Scene7, distanceMultiplier,n, blend);
  AddClones(Scene8, distanceMultiplier,n, blend);
  AddClones(Scene9, distanceMultiplier,n, blend);
  AddClones(Scene10, distanceMultiplier,n, blend);
end;

procedure TViewMain.TestImages(distanceMultiplier, ANumber: Integer; blend: Boolean);
begin
  Visibility_Planes(false, blend);
  Visibility_Scenes(false, blend);
  Visibility_Images(true, blend);

  AddImages(TreeImage, distanceMultiplier, ANumber-1, blend);
end;

procedure TViewMain.TestPlanes(distanceMultiplier, ANumber: Integer; blend: Boolean);
begin
  Visibility_Images(false, blend);
  Visibility_Scenes(false, blend);
  Visibility_Planes(true, blend);

  AddPlanes(Plane1, distanceMultiplier, ANumber-1, blend);
end;

procedure TViewMain.TestScenes(distanceMultiplier, ANumber: Integer; blend: Boolean);
var
  n: Integer;
begin
  Visibility_Images(false, blend);
  Visibility_Planes(false, blend);
  Visibility_Scenes(true, blend);

  n := (ANumber div 10) - 1;
  AddScenes(Scene1, distanceMultiplier,n, blend);
  AddScenes(Scene2, distanceMultiplier,n, blend);
  AddScenes(Scene3, distanceMultiplier,n, blend);
  AddScenes(Scene4, distanceMultiplier,n, blend);
  AddScenes(Scene5, distanceMultiplier,n, blend);
  AddScenes(Scene6, distanceMultiplier,n, blend);
  AddScenes(Scene7, distanceMultiplier,n, blend);
  AddScenes(Scene8, distanceMultiplier,n, blend);
  AddScenes(Scene9, distanceMultiplier,n, blend);
  AddScenes(Scene10, distanceMultiplier,n, blend);
end;

procedure TViewMain.Update(const SecondsPassed: Single; var HandleInput: Boolean);
begin
  inherited;
  { This virtual method is executed every frame (many times per second). }
  Assert(LabelFps <> nil, 'If you remove LabelFps from the design, remember to remove also the assignment "LabelFps.Caption := ..." from code');
  LabelFps.Caption := 'FPS: ' + Container.Fps.ToString;
end;

function TViewMain.Press(const Event: TInputPressRelease): Boolean;
var
  k,n: Integer;
  blend: Boolean;
  whatToTest: TDiggiTestCase;
begin
  Result := inherited;
  if Result then Exit; // allow the ancestor to handle keys

  {
    Various PNG images, all 1024px high, width kept around 700px
    with alpha channel & without it (custom color = transparent color)

    Blending is: acBlending for images, plus RenderOptions.Blending for all objects

    Viewport blending sort: sortAuto
    Release mode
    Compiler: Delphi, Windows 11
    AlphaChannel  # trees  Fps:
                           Image  Plane  Scene  Clone  Clone*10
    acNone        2000     30     22     27     30     30
    acTest        2000     27
    Blending      2000     42     37     40     41     44

    acNone        5000     9      8      9      11.8   12
    acTest        5000     8.3
    Blending      5000     13     11     13     18     17

    Compiler: Lazarus, Windows 11
    AlphaChannel  # trees  Fps:
                           Image  Plane  Scene  Clone  Clone*10
    acNone        2000     63     35     45     60       61
    acTest        2000     50
    Blending      2000     72     47     63     80     80

    acNone        5000     15     10     12     25     24
    acTest        5000     13
    Blending      5000     20     15     18     31.5   32
  }

  if Event.IsKey(keyF1) then
  begin
    k := 4;
    n := 2000;
    blend := false;
    whatToTest := diggiTestClones;

    if whatToTest = diggiTestImages then
      TestImages(k,n,blend);
    if whatToTest = diggiTestPlanes then
      TestPlanes(k,n,blend);
    if whatToTest = diggiTestScenes then
      TestScenes(k,n,blend);
    if whatToTest = diggiTestClones then
      TestClones(k,n,blend);

    Exit(true); // key was handled
  end;
end;

end.
